library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use work.const.all;



entity id_stage is
port (
	clk: in std_logic;
	reset: in std_logic;
	stall, nop: in std_logic;

	i_pc: in adr_type;
	ir_in: in dat_type;
	i_exc_vector: in exc_vector_type;
	i_isValid: in std_logic;

	--from write back
	wb_rdDat: in dat_type;
	wb_rdAdr: in reg_adr_type;
	wb_ena: in std_logic;

	--operation type
	o_op_alu: out std_logic_vector(3 downto 0);
	--operand width -- for now only for MEM operations
	o_sel8:  out std_logic;
	o_sel16: out std_logic;
	o_sel32: out std_logic;
	--instruction type
	o_isValid: out std_logic;
	o_isMem: out std_logic;
	o_isLdi: out std_logic;
	o_isAlu: out std_logic;
	o_isHigh: out std_logic;
	o_isLow: out std_logic;
	o_isStore: out std_logic;
	o_isBranch: out std_logic;
	o_isAluSpec: out std_logic;
	o_isHalt: out std_logic;
	--
	o_rdAdr: out reg_adr_type;
	o_rs1Adr: out reg_adr_type;
	o_rs2Adr: out reg_adr_type;
	--
	o_imm: out dat_type;

	--register operands
	o_rs1: out dat_type;
	o_rs2: out dat_type;

	o_isJmp_if: out std_logic; -- go to if directly, not through latches
	o_jmpAdr: out adr_type;
	o_pc, o_branch_adr: out adr_type;
	o_branch_cond: out std_logic_vector(1 downto 0);
	
	--stalling due to LOAD-ALU hazard
	id_isMem, id_isStore: in std_logic;
	id_rdAdr: in reg_adr_type;
	--to stall_nop_control
	o_load_alu_stall: out std_logic;

	o_exc_vector: out exc_vector_type;
	i_rmPC, i_rmAddress: in adr_type; --from WB
	i_rm_ena: in std_logic; --from WB
	i_trapPC_ena: std_logic;

	o_trapPC: out adr_type;

	--itlb_dat, itlb_adr: out dat_type;
	o_itlb_ena: out std_logic
);
end entity id_stage;

architecture behav of id_stage is
--from decoder
	signal i_op_alu: std_logic_vector(3 downto 0);
	signal i_sel8, i_sel16, i_sel32 : std_logic;
	signal i_isHigh, i_isLow, i_isMem, i_isLdi, i_isAlu, i_isAluSpec: std_logic;
	signal i_isJmp, isJmpSpec, i_isStore, i_isBranch, i_isHalt: std_logic;
	signal i_rdAdr, i_rs1Adr, i_rs2Adr: reg_adr_type;
	signal i_imm: dat_type;
	signal i_branch_cond: std_logic_vector(1 downto 0);
--from reg file
	signal i_rs1_tmp, i_rs2 : dat_type;
	signal i_rs1: dat_type;

	signal i_branch_adr, jmp_addr: adr_type;

--stall LOAD-ALU
	signal reg_stall, rf_ena, load_alu_stall: std_logic;
--exceptions
	signal exc_vector: exc_vector_type;
	signal exc_decode: std_logic;
	signal rmPC, rmAddress, trapPC: adr_type;
	signal rmLock: std_logic;
	
	signal itlb_ena: std_logic;
	signal wr_regs: std_logic;
begin
	decoder: entity work.id_decoder(behav)
		PORT MAP(
			ir_in,

			i_op_alu,
			i_sel8, i_sel16, i_sel32,
			i_isMem, i_isLdi, i_isAlu, i_isJmp, isJmpSpec,
			i_isHigh, i_isLow, i_isStore, i_isBranch, i_isAluSpec, i_isHalt,

			i_rdAdr, i_rs1Adr, i_rs2Adr,
			i_imm, i_branch_cond, exc_decode
		);

	o_isJmp_if <= i_isJmp;
	i_branch_adr <= i_pc + (i_imm(29 downto 0) & "00");
	o_jmpAdr <= i_imm(29 downto 0) & "00" when isJmpSpec = '0' else
				rmPC;

-- LOAD-ALU stall
	load_alu_stall <= 	'1' when id_isMem = '1' and id_isStore = '0' and
								(i_rs1Adr = id_rdAdr or i_rs2Adr = id_rdAdr) else
						'0'; 
    o_load_alu_stall <= load_alu_stall;

-- write to iTLB
	itlb_ena <= '1' when ir_in(adr_w-1 downto adr_w-8) = "01000010" else '0'; --make this more beautiful

    rmRegs: process(clk, reset, i_rm_ena, isJmpSpec, rmLock) 
    begin
    	if isJmpSpec = '1' then
    		rmLock <= '0';
    	elsif reset = '1' then
    		rmPC <= (others => '0');
    		rmAddress <= (others => '0');
    		rmLock <= '0';
    	elsif clk = '1' and clk'event and i_rm_ena = '1' and rmLock = '0' then
    		rmPC <= i_rmPC;
    		rmAddress <= i_rmAddress;
    		rmLock <= '1';
    	end if;
    end process rmRegs;

    trapPcUpdate: process(clk, reset, i_trapPC_ena)
    begin
    	if reset = '1' then
    		trapPC <= (others => '0');
    	elsif (clk = '1' and clk'event and i_trapPC_ena = '1') then
    		trapPC <= wb_rdDat;
    	end if;
    end process trapPcUpdate;
    o_trapPC <= trapPC;

	wr_regs <= wb_ena and not stall;
	reg_file: entity work.id_reg_file(behav)
		PORT MAP(
			clk,
			reset,

			wr_regs,
			i_rs1Adr, i_rs2Adr,
			wb_rdAdr, wb_rdDat, i_rs1_tmp, i_rs2
		);

	i_rs1 <= rmAddress when i_isAluSpec = '1' and i_branch_cond = "10" else
			 i_rs1_tmp;


	id_inj_opcode: entity work.exc_inject(behav)
		GENERIC MAP(exc_id_opcode)
		PORT MAP(
			exc_decode,
			i_exc_vector,
			exc_vector
		);


	latch: entity work.id_latch(behav)
		PORT MAP(
			clk, reset, stall, nop,

			i_op_alu,
			i_sel8, i_sel16, i_sel32,
			i_isValid, i_isMem, i_isLdi, i_isAlu, i_isAluSpec,
			i_isHigh, i_isLow, i_isStore, i_isBranch, i_isHalt,
			i_rdAdr, i_rs1Adr, i_rs2Adr,
			i_imm, i_branch_cond, i_pc, i_branch_adr,
			i_rs1, i_rs2,
			exc_vector,
			itlb_ena,

			-- output signals
			o_op_alu,
			o_sel8, o_sel16, o_sel32,
			o_isValid, o_isMem, o_isLdi, o_isAlu, o_isAluSpec,
			o_isHigh, o_isLow, o_isStore, o_isBranch, o_isHalt,
			o_rdAdr, o_rs1Adr, o_rs2Adr,
			o_imm, o_branch_cond, o_pc, o_branch_adr,
			o_rs1, o_rs2,
			o_exc_vector,
			o_itlb_ena
		);


end architecture behav;
