library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use work.const.all;


entity mem_stage is
port (
	clk: in std_logic;
	reset: in std_logic;
	stall, nop: in std_logic;

	i_pc: in adr_type;
	i_alu_data: in adr_type;
	i_datain: in dat_type;
	o_dataout: out dat_type;

	i_branch_cond: in std_logic_vector(1 downto 0);
	sel8, sel16, sel32: in std_logic;
	i_isValid, i_ismem, i_ldi: in std_logic;
	i_isalu, i_isaluspec, i_isstore, i_ishalt: in std_logic;
	i_reg_addr: in reg_adr_type;
	i_regWrEna: in std_logic;
	i_rs2Adr: in reg_adr_type;
	i_exc_vector: in exc_vector_type;

	o_pc: out adr_type;
	o_branch_cond: out std_logic_vector(1 downto 0);
	o_isValid, o_ismem, o_ldi: out  std_logic;
	o_isalu, o_isaluspec, o_isstore, o_ishalt: out  std_logic;
	o_reg_addr: out reg_adr_type;
	o_alu_data: out dat_type;
	o_regWrEna: out std_logic;

	o_exc_vector: out exc_vector_type;
	o_rmAddress: out adr_type;

	--bypass MEM->MEM
	i_mem_ismem, i_mem_isstore: in std_logic;
	i_mem_rdAdr: in reg_adr_type;
	i_mem_dataout: in dat_type;

	--cache
	line_addr: out adr_type;
	line_idata: in line_dat_type;
	line_odata: out line_dat_type;
	miss: out std_logic;
	o_bfull: out std_logic;
	fill: in std_logic;
	mem_rd, mem_wr: out std_logic
);
end entity mem_stage;

architecture behav of mem_stage is
	signal cache_out, datain: dat_type;
	signal exc_vector: exc_vector_type;
	signal rd, wr: std_logic;
	signal exc_align: std_logic;
	signal rmAddress, tlb_adr, phys_adr: adr_type;
	signal tlb_data: dat_type;
	signal tlb_wrEna: std_logic;
	signal exc_vector_align, exc_vector_access: exc_vector_type;
	signal exc_cache_align, exc_dtlb_access, exc_dtlb_miss: std_logic;
	signal dtlb_access, dtlb_miss: std_logic;
	signal cache_wr, cache_miss: std_logic;
	signal buf_wr, buf_shift, buf_empty, buf_full: std_logic;
	signal bfull, drain_buf: std_logic;
	signal buf_hit: std_logic_vector(3 downto 0);
	signal buf_addr, vir_cache_addr, phy_cache_addr: adr_type;
	signal buf_data: dat_type;
begin
	rd <= i_ismem and not i_isstore;
	wr <= i_ismem and i_isstore;

	--bypass MEM->MEM
	datain <= i_mem_dataout when i_mem_ismem = '1' and
	                             i_mem_isstore = '0' and
	                             i_isstore = '1' and
	                             i_mem_rdAdr = i_rs2Adr else
	          i_datain;

	exc_cache_align <= exc_align and i_ismem;
	exc_inj_align: entity work.exc_inject(behav)
		GENERIC MAP(exc_mem_align)
		PORT MAP(
			exc_cache_align,
			i_exc_vector,
			exc_vector_align
		);

	exc_dtlb_access <= dtlb_access and not exc_align and i_ismem;
	exc_inj_access: entity work.exc_inject(behav)
		GENERIC MAP(exc_mem_access)
		PORT MAP(
			exc_dtlb_access,
			exc_vector_align,
			exc_vector_access
		);

	exc_dtlb_miss <= dtlb_miss and not exc_dtlb_access and i_ismem;
	exc_inj_miss: entity work.exc_inject(behav)
		GENERIC MAP(exc_mem_dtlb)
		PORT MAP(
			exc_dtlb_miss,
			exc_vector_access,
			exc_vector
		);

	tlb_data <= i_alu_data;
	tlb_adr <= datain;
	tlb_wrEna <= '1' when i_isalu = '1' and i_branch_cond = "11" else
				 '0'; --assumption: ins(25) = '1' only for TLB instructions in ALU group
	dtlb: entity work.tlb(behav)
		PORT MAP(
			clk, reset,
			i_alu_data,
			rd, wr, '0',
			tlb_data,
			tlb_adr,
			tlb_wrEna,
			phys_adr,
			dtlb_miss, dtlb_access
		);

	phy_cache_addr <= buf_addr when cache_wr = '1' else
	                  phys_adr;
	vir_cache_addr <= buf_addr when cache_wr = '1' else
	                  i_alu_data;
	cache: entity work.dcache(rtl)
		PORT MAP(clk, reset,
		         i_ismem, rd, cache_wr,
		         sel8, sel16, sel32,
		         vir_cache_addr,
		         -- data output from cache
		         cache_out,
		         -- from TLB
		         phy_cache_addr,
		         -- store buffer signals
		         buf_data, buf_hit, buf_empty, drain_buf,
		         -- lines from/to memory
		         line_idata, line_odata,
		         -- control signals
		         cache_miss, fill, line_addr, mem_rd, mem_wr,
		         exc_align, exc_dtlb_miss);
	miss <= cache_miss;

	buf_shift <= (not i_ismem and not buf_empty) or drain_buf or bfull;
	buf_wr <= wr and not cache_miss and not stall and not exc_dtlb_miss;
	bfull <= buf_full and wr and not cache_miss;
	o_bfull <= bfull;
	sbuff: entity work.storeb(rtl)
		PORT MAP(clk, reset,
		         buf_wr, buf_shift,
		         sel8, sel16, sel32,
		         datain, phys_adr,
		         --data out
		         buf_addr, buf_data, buf_hit,
		         -- control signals
		         buf_empty, buf_full, cache_wr);


	--rmAddress = PC if there was iTLB miss
	--otherwise virtual address that goes into TLB
	rmAddress <= i_pc when i_exc_vector(exc_if_itlb) = '1' else
	             i_alu_data;
	MEM_LATCH: entity work.mem_latch(behav)
		PORT MAP(clk, reset, stall, nop,
			cache_out, o_dataout,

			i_pc,
			i_branch_cond,
			i_isValid, i_ismem, i_ldi,
			i_isalu, i_isaluspec, i_isstore, i_ishalt,
			i_reg_addr, i_alu_data,
			i_regWrEna,
			exc_vector,
			rmAddress,

			o_pc,
			o_branch_cond,
			o_isValid, o_ismem, o_ldi,
			o_isalu, o_isaluspec, o_isstore, o_ishalt,
			o_reg_addr, o_alu_data,
			o_regWrEna,
			o_exc_vector,
			o_rmAddress
		);

end architecture behav;
