
#include <errno.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "as.h"

struct label labels[MAXLABELS], *curlabel = labels;
struct code codes[MAXCODES], *curcode = codes;
jmp_buf recovery;

static int iserror, good_end;
static char *finput, *foutput, *folabel, *filabel;

void
error_(unsigned linenum, char *fmt, va_list va)
{
	int c;
	extern int iserror;

	fputs(PROG ": ", stderr);
	fprintf(stderr, "%s:%u:", finput, linenum);
	vfprintf(stderr, fmt, va);
	putc('\n', stderr);
	iserror = 1;
	longjmp(recovery, 1);
}

void
die(char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	fputs(PROG ": ", stderr);
	vfprintf(stderr, fmt, va);
	putc('\n', stderr);
	va_end(va);
	exit(EXIT_FAILURE);
}

static void
usage(void)
{
	fputs(PROG " [-h][-o output.bin][-l [file.lst]][-i [file.lst]][input.as]\n"
	      "\t-h\tShow this help\n"
	      "\t-o\tOutput file\n"
	      "\t-l\tOutput listing file (labels.lst by default)\n"
	      "\t-i\tInput listing file (labels.lst by default)\n"
	      "\tinput\tInput file\n", stderr);
	exit(EXIT_FAILURE);
}

static void
cleanup(void)
{
	if (!good_end) {
		if (foutput)
			remove(foutput);
		if (folabel)
			remove(folabel);
	}
}

static int
rlabels(char *fname)
{
	char buf[MAXLABEL_SIZ];
	long val;
	struct label *lp;
	FILE *fp;

	if (fname[0] == '-' && fname[1] == '\0')
		fp = stdin;
	else
		fp = fopen(fname, "r");

	while (fscanf(fp, "%8[A-Za-z0-9]:\t%0lX ", buf, &val) == 2) {
		lp = lookup(buf);
		if (lp->def) {
			fclose(fp);
			return -1;
		}
		lp->def = 1;
		lp->offset = val;
	}

	return !feof(fp) | ferror(fp) | fclose(fp);
}

static int
plabels(char *fname)
{
	struct label *lp;
	FILE *fp;

	if (fname[0] == '-' && fname[1] == '\0')
		fp = stdout;
	else
		fp = fopen(fname, "w");

	if (!fp)
		return -1;
	for (lp = labels; lp < curlabel; ++lp)
		fprintf(fp, "%s:\t%08X\n", lp->s, lp->offset);
	return ferror(fp) | fclose(fp);
}

static void
redirect(void)
{
	/* +1 because s/.as/.bin/ generates one character. In this case the
	 * name will be no valid and then there will be an error in freopen.
	 * It's safe ;)
	 */
	static char buf[FILENAME_MAX+1];

	if (finput) {
		char *p = strrchr(finput, '.');

		if (!p || strcmp(p, ".as"))
			die("incorrect input file name '%s'", finput);
		if (!freopen(finput, "r", stdin)) {
			die("error opening input file '%s':%s",
			    finput, strerror(errno));
		}
		if (!foutput) {
			size_t n = p - finput;
			memcpy(buf, finput, n);
			strcpy(buf + n, ".bin");
			foutput = buf;
		}
	} else {
		finput = "<stdin>";
	}

	if (foutput && strcmp(foutput, "-")) {
		if (!freopen(foutput, "w", stdout)) {
			die("error opening output file '%s':%s",
			    foutput, strerror(errno));
		}
		atexit(cleanup);
	} else {
		foutput = "<stdout>";
	}
}

int
main(int argc, char *argv[])
{
	char c, *s;

	while (*++argv) {
		if (argv[0][0] == '-') {
			c = argv[0][1];
			if (!(s = *++argv) ||  s[0] == '-' && s[1] != '\0')
				s = NULL;
			switch (c) {
			case 'o':
				if (foutput)
					goto duplicated;
				if (!(foutput = s))
					goto need_parameter;
				break;
			case 'l':
				if (folabel)
					goto duplicated;
				if (!s) {
					--argv;
					s = "labels.lst";
				}
				folabel = s;
				break;
			case 'i':
				if (filabel)
					goto duplicated;
				if (!s) {
					--argv;
					s = "labels.lst";
				}
				filabel = s;
				break;
			default:
				fprintf(stderr,
				        "incorrect parameter '%c'\n", c);
			case 'h':
				usage();
			duplicated:
				die("duplicated -%c", c);
			need_parameter:
				die("-%c needs a parameter", c);
			}
		} else {
			if (finput)
				die("more of one input file");
			finput = *argv;
		}
	}

	redirect();
	if (filabel && rlabels(filabel))
		die("error reading input labels from '%s'", filabel);
	pass1();
	if (ferror(stdin))
		die("error reading from '%s':%s", strerror(errno));
	if (iserror)
		exit(EXIT_FAILURE);

	pass2();
	if (ferror(stdout))
		die("error reading from '%s':%s", strerror(errno));
	if (iserror)
		exit(EXIT_FAILURE);
	if (folabel && plabels(folabel))
		die("error writing output labels in '%s'", folabel);
	good_end = 1;
	return 0;
}
